//=============================================================================
// CastleDB.js
//=============================================================================

/*:ja
 * @plugindesc CastleDBのデータベース参照機能を追加します。
 * @author RaTTiE
 * 
 * @param Auto Load
 * @desc 自動ロード機能の有効状態です。有効時は$cdbへ自動的にロード処理が行われます。
 * @default true
 *
 * @param Data Path
 * @desc cdbファイルの格納先です。
 * @default ./data/Data.cdb
 * 
 * @param Image Path
 * @desc imgファイルの格納先です。
 * @default ./data/Data.img
 *
 * @help
 * 関数リファレンスはREADMEを参照してください。
 */
(function() {
	var TYPE_STR_UNIQUE_IDENTIFIER = '0';
	var TYPE_STR_ENUM = '5';
	var TYPE_STR_FLAG = '10';
	var TYPE_STR_REFERENCE = '6';
	var TYPE_STR_IMAGE = '7';
	var TYPE_STR_LIST = '8';
	var TYPE_STR_COLOR = '11';
	var TYPE_STR_CUSTOM_TYPE = '9';

	var parameters = PluginManager.parameters('CastleDB');
	var dataPath = String(parameters['Data Path']) || './data/Data.cdb';
	var imagePath = String(parameters['Image Path']) || './data/Data.img';
	var autoLoad = parameters['Auto Load'] == 'true';

	// Scene_Bootの改変
	var orgScene_Boot_create = Scene_Boot.prototype.create;
	Scene_Boot.prototype.create = function() {
		orgScene_Boot_create.call(this);

		if (autoLoad) {
			window.$cdb = CastleDBManager.load(dataPath, imagePath);
		}
	};

	var orgScene_Boot_isReady = Scene_Boot.prototype.isReady;
	Scene_Boot.prototype.isReady = function() {
		if (!autoLoad) {
			return orgScene_Boot_isReady.call(this);
		} else if (orgScene_Boot_isReady.call(this)) {
			return $cdb.isDatabaseLoaded();
		} else {
			return false;
		}
	}

	// ImageManagerの改変
	ImageManager.loadCastleDBBitmap = function(url, md5, src, hue) {
		var key = url + ':' + md5 + ':' + hue;
		var bitmap = this._cache[key];
		
		if (!bitmap) {
			var bitmap = new Bitmap();

			bitmap._image = new Image();
			bitmap._url = url || imagePath;
			bitmap._isLoading = true;
			bitmap._image.src = src;
			bitmap._image.onload = Bitmap.prototype._onLoad.bind(bitmap);
			bitmap._image.onerror = Bitmap.prototype._onError.bind(bitmap);

			bitmap.addLoadListener(function() {
				bitmap.rotateHue(hue);
			});

			this._cache[key] = bitmap;
		}

		return bitmap;
	};

	// DBManagerの実装
	function CastleDBManager() {
		this.initialize.apply(this, arguments);
	}

	CastleDBManager.prototype.initialize = function() {
		this._dataPath = null;
		this._imagePath = null;

		this.reset();
	};

	CastleDBManager.prototype.reset = function() {
		this._dataErrorUrl = null;
		this._imageErrorUrl = null;

		this.resetData();
		this.resetImage();
	}

	CastleDBManager.prototype.resetData = function() {
		this._data = null;
		this._sheetCache = {};
		this._sheetSchemaCache = {};
		this._sheetIdentifierCache = {};
		this._sheetIdentifyCache = {};
	}

	CastleDBManager.prototype.resetImage = function() {
		this._image = null;
	}

	CastleDBManager.load = function(dataPath, imagePath) {
		var database = new CastleDBManager();
		database._dataPath = dataPath;
		database._imagePath = imagePath;

		database.loadDataFile();
		database.loadImageFile();

		return database;
	}

	CastleDBManager.prototype.loadDataFile = function() {
		var database = this;
		database.resetData();

		var xhr = new XMLHttpRequest();
		var url = database._dataPath;
		xhr.open('GET', url);
		xhr.overrideMimeType('applidation/json');

		xhr.onload = function() {
			if (xhr.status < 400) {
				database._data = JSON.parse(xhr.responseText);
			}
		};
		xhr.onerror = function() {
			database._dataErrorUrl = database._dataErrorUrl || url;
		};

		xhr.send();
	};

	CastleDBManager.prototype.loadImageFile = function() {
		var database = this;
		database.resetImage();

		if (!database._imagePath) return;

		var xhr = new XMLHttpRequest();
		var url = database._imagePath;
		xhr.open('GET', url);
		xhr.overrideMimeType('applidation/json');

		xhr.onload = function() {
			if (xhr.status < 400) {
				database._image = JSON.parse(xhr.responseText);
			}
		};
		xhr.onerror = function() {
			database._imageErrorUrl = database._imageErrorUrl || url;
		};

		xhr.send();
	};

	CastleDBManager.prototype.isDatabaseLoaded = function() {
		this.checkError(false);

		if (!this._data) return false;
		if (this._imagePath && !this._image && !this._imageErrorUrl) return false;

		return true;
	};

	CastleDBManager.prototype.checkError = function(checkImageError) {
		if (this._dataErrorUrl) {
			throw new Error('Failed to load: ' + this._dataErrorUrl);
		}

		if (checkImageError && this._imageErrorUrl) {
			throw new Error('Failed to load: ' + this._imageErrorUrl);
		}
	};

	CastleDBManager.prototype.getSheet = function(sheetName) {
		if (!this._sheetCache[sheetName]) {
			for (var i = 0; i < this._data.sheets.length; i++) {
				if (this._data.sheets[i].name == sheetName) {
					this._sheetCache[sheetName] = this._data.sheets[i];

					break;
				}
			}

			if (!this._sheetCache[sheetName]) {
				throw new Error('Sheet not found: ' + sheetName);
			}
		}

		return this._sheetCache[sheetName];
	};

	CastleDBManager.prototype.getSchema = function(sheetName, columnName) {
		assertParameters(arguments, ['sheetName', 'columnName']);

		if (!this._sheetSchemaCache[sheetName]) {
			var schema = {};
			var columns = this.getSheet(sheetName).columns;

			for (var i = 0; i < columns.length; i++) {
				var column = columns[i];

				schema[column.name] = column;
			}

			this._sheetSchemaCache[sheetName] = schema;
		}

		var column = this._sheetSchemaCache[sheetName][columnName];

		if (column !== undefined) {
			return column;
		} else {
			throw new Error('Column not found: ' + sheetName + '/' + columnName);
		}
	}

	CastleDBManager.prototype.getIdentifier = function(sheetName) {
		assertParameters(arguments, ['sheetName']);

		if (!this._sheetIdentifierCache[sheetName]) {
			var columns = this.getSheet(sheetName).columns;

			for (var i = 0; i < columns.length; i++) {
				var column = columns[i];

				if (column.typeStr == TYPE_STR_UNIQUE_IDENTIFIER) {
					this._sheetIdentifierCache[sheetName] = column.name;

					break;
				}
			}
		}

		var identifier = this._sheetIdentifierCache[sheetName];
		if (identifier !== undefined) {
			return identifier;
		} else {
			throw new Error('Unique Identifier undefined: ' + sheetName);
		}
	};

	CastleDBManager.prototype.getLines = function(sheetName) {
		assertParameters(arguments, ['sheetName']);

		return this.getSheet(sheetName).lines;
	}

	CastleDBManager.prototype.getLine = function(sheetName, id) {
		assertParameters(arguments, ['sheetName', 'id']);

		return this.getItem(sheetName, id).getLine();
	}

	CastleDBManager.prototype.getValue = function(sheetName, id, columnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getValue(columnName);
	};

	CastleDBManager.prototype.getItemInner = function(sheetName, lines, id, sheetIdentifyCache) {
		assertParameters(arguments, ['sheetName', 'lines', 'id', 'sheetIdentifyCache']);

		var manager = this;
		var createItem = function(line, index) {
			return new CastleDBManagerItem(manager, sheetName, line, index);
		};

		var item;
		if (typeof id === 'string' || id instanceof String) {
			if (sheetIdentifyCache == null) sheetIdentifyCache = {};

			if (!sheetIdentifyCache[sheetName]) {
				sheetIdentifyCache[sheetName] = this.getIdentifyItems(sheetName, lines);
			}

			item = sheetIdentifyCache[sheetName][id];
		} else {
			var isInteger = (
				(typeof id === 'number' || id instanceof Number) &&
				id == parseInt(id, 10)
			);

			if (isInteger) {
				item = new CastleDBManagerItem(manager, sheetName, lines[id], id);
			} else {
				throw new Error('Bad identifier: ' + id);
			}
		}

		if (item !== undefined) {
			return item;
		} else {
			throw new Error('Item not found: ' + sheetName + '/' + id);
		}
	};

	CastleDBManager.prototype.getIdentifyItems = function(sheetName, lines) {
		assertParameters(arguments, ['sheetName', 'lines']);

		var items = {};
		var identifier = this.getIdentifier(sheetName);

		for (var i = 0; i < lines.length; i++) {
			var line = lines[i];

			items[line[identifier]] = new CastleDBManagerItem(this, sheetName, line, i);
		}

		return items;
	};

	CastleDBManager.prototype.getItem = function(sheetName, id) {
		assertParameters(arguments, ['sheetName', 'id']);

		return this.getItemInner(
			sheetName,
			this.getLines(sheetName),
			id,
			this._sheetIdentifyCache
		);
	};

	CastleDBManager.prototype.getEnumDefinition = function(sheetName, columnName) {
		assertParameters(arguments, ['sheetName', 'columnName']);

		var column = this.getSchema(sheetName, columnName);

		if (column.typeStr.indexOf(TYPE_STR_ENUM + ':') != 0) {
			throw new Error('Column is not enum: ' + sheetName + '/' + columnName);
		}

		return column.typeStr.substring((TYPE_STR_ENUM + ':').length).split(',');
	};

	CastleDBManager.prototype.getFlagDefinition = function(sheetName, columnName) {
		assertParameters(arguments, ['sheetName', 'columnName']);

		var column = this.getSchema(sheetName, columnName);

		if (column.typeStr.indexOf(TYPE_STR_FLAG + ':') != 0) {
			throw new Error('Column is not flag: ' + sheetName + '/' + columnName);
		}

		return column.typeStr.substring((TYPE_STR_FLAG + ':').length).split(',');
	};

	CastleDBManager.prototype.getCustomTypeDefinition = function(sheetName, columnName) {
		assertParameters(arguments, ['sheetName', 'columnName']);

		var column = this.getSchema(sheetName, columnName);

		if (column.typeStr.indexOf(TYPE_STR_CUSTOM_TYPE + ':') != 0) {
			throw new Error('Column is not custom type: ' + sheetName + '/' + columnName);
		}

		var customTypeName = column.typeStr.substring((TYPE_STR_CUSTOM_TYPE + ':').length);
		var customTypes = this._data.customTypes;

		for (var i = 0; i < customTypes.length; i++) {
			var customType = customTypes[i];

			if (customType.name == customTypeName) {
				return customType.cases;
			}
		}

		throw new Error('Custom type is not found: ' + customType);
	};

	CastleDBManager.prototype.getEnumValue = function(sheetName, id, columnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getEnumValue(columnName);
	};

	CastleDBManager.prototype.getFlagValues = function(sheetName, id, columnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getFlagValues(columnName);
	};

	CastleDBManager.prototype.getReferenceItem = function(sheetName, id, columnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getReferenceItem(columnName);
	};

	CastleDBManager.prototype.getReferenceValue = function(sheetName, id, columnName, refColumnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getReferenceValue(columnName, refColumnName);
	};

	CastleDBManager.prototype.getListLines = function(sheetName, id, columnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getListLines(columnName);
	};

	CastleDBManager.prototype.getListItems = function(sheetName, id, columnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getListItems(columnName);
	};

	CastleDBManager.prototype.getListItem = function(sheetName, id, columnName, listId) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName', 'listId']);

		return this.getItem(sheetName, id).getListItem(columnName, listId);
	};

	CastleDBManager.prototype.getListValue = function(sheetName, id, columnName, listId, listColumnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName', 'listId', 'listColumnName']);

		return this.getListItem(sheetName, id, columnName, listId).getValue(listColumnName);
	};

	CastleDBManager.prototype.getListIdentifyItems = function(sheetName, id, columnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getListIdentifyItems(columnName);
	};

	CastleDBManager.prototype.getColor = function(sheetName, id, columnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getColor(columnName);
	};

	CastleDBManager.prototype.getCustomType = function(sheetName, id, columnName) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getCustomType(columnName);
	};

	CastleDBManager.prototype.getImage = function(sheetName, id, columnName, hue) {
		assertParameters(arguments, ['sheetName', 'id', 'columnName']);

		return this.getItem(sheetName, id).getImage(columnName, hue);
	};

	// アイテム定義
	function CastleDBManagerItem(manager, sheetName, line) {
		this.initialize.apply(this, arguments);
	}

	CastleDBManagerItem.prototype.initialize = function(manager, sheetName, line, listIndex) {
		this._manager = manager;
		this._sheetName = sheetName;
		this._line = line;
		this._listIndex = listIndex;
	};

	CastleDBManagerItem.prototype.getLine = function() {
		return this._line;
	};

	CastleDBManagerItem.prototype.getValue = function(columnName) {
		assertParameters(arguments, ['columnName']);

		var value = this._line[columnName];
		if (value === undefined) {
			this._manager.getSchema(this._sheetName, columnName);

			value = null;
		}

		return value;
	};

	CastleDBManagerItem.prototype.getEnumDefinition = function(columnName) {
		assertParameters(arguments, ['columnName']);

		return this._manager.getEnumDefinition(this._sheetName, columnName);
	};

	CastleDBManagerItem.prototype.getFlagDefinition = function(columnName) {
		assertParameters(arguments, ['columnName']);

		return this._manager.getFlagDefinition(this._sheetName, columnName);
	};

	CastleDBManagerItem.prototype.getCustomTypeDefinition = function(columnName) {
		assertParameters(arguments, ['columnName']);

		return this._manager.getCustomTypeDefinition(this._sheetName, columnName);
	};

	CastleDBManagerItem.prototype.getSchema = function(columnName) {
		assertParameters(arguments, ['columnName']);

		return this._manager.getSchema(this._sheetName, columnName);
	};

	CastleDBManagerItem.prototype.getEnumValue = function(columnName) {
		assertParameters(arguments, ['columnName']);

		return this.getEnumDefinition(columnName)[this.getValue(columnName)];
	};

	CastleDBManagerItem.prototype.getFlagValues = function(columnName) {
		assertParameters(arguments, ['columnName']);

		var values = [];
		var definitions = this.getFlagDefinition(columnName);

		for (var i = 0; i < definitions.length; i++) {
			if (this.getValue(columnName) & Math.pow(2, i)) {
				values.push(definitions[i]);
			}
		}

		return values;
	};

	CastleDBManagerItem.prototype.getReferenceItem = function(columnName) {
		assertParameters(arguments, ['columnName']);

		var column = this.getSchema(columnName);

		if (column.typeStr.indexOf(TYPE_STR_REFERENCE + ':') != 0) {
			var address = sheetName + '/' + columnName;
			if (this._listIndex !== undefined) {
				address += '(' + this._listIndex + ')';
			}

			throw new Error('Column is not reference: ' + address);
		}

		var referenceSheetName = column.typeStr.substring((TYPE_STR_REFERENCE + ':').length);

		return this._manager.getItem(referenceSheetName, this.getValue(columnName));
	};

	CastleDBManagerItem.prototype.getReferenceValue = function(columnName, refColumnName) {
		assertParameters(arguments, ['columnName', 'refColumnName']);

		return this.getReferenceItem(columnName).getValue(refColumnName);
	};

	CastleDBManagerItem.prototype.getListLines = function(columnName) {
		assertParameters(arguments, ['columnName']);

		var column = this.getSchema(columnName);

		if (column.typeStr != TYPE_STR_LIST) {
			throw new Error('Column is not list: ' + sheetName + '/' + columnName);
		}

		return this.getValue(columnName);
	};

	CastleDBManagerItem.prototype.getListSheetName = function(columnName) {
		assertParameters(arguments, ['columnName']);

		return this._sheetName + '@' + columnName;
	}

	CastleDBManagerItem.prototype.getListItem = function(columnName, id) {
		assertParameters(arguments, ['columnName', 'id']);

		return this._manager.getItemInner(
			this.getListSheetName(columnName),
			this.getListLines(columnName),
			id,
			this._manager._sheetIdentifyCache
		);
	};

	CastleDBManagerItem.prototype.getListValue = function(columnName, id, listColumnName) {
		assertParameters(arguments, ['columnName', 'id', 'listColumnName']);

		return this.getListItem.getValue(listColumnName);
	};

	CastleDBManagerItem.prototype.getListItems = function(columnName) {
		assertParameters(arguments, ['columnName']);

		var list = [];
		var lines = this.getListLines(columnName);

		for (var i = 0; i < lines.length; i++) {
			list.push(
				new CastleDBManagerItem(
					this._manager,
					this.getListSheetName(columnName),
					lines[i],
					i
				)
			);
		}

		return list;
	};

	CastleDBManagerItem.prototype.getListIdentifyItems = function(columnName) {
		assertParameters(arguments, ['columnName']);

		return this._manager.getIdentifyItems(
			this.getListSheetName(columnName),
			this.getListLines(columnName)
		);
	};

	CastleDBManagerItem.prototype.getColor = function(columnName) {
		assertParameters(arguments, ['columnName']);

		var column = this.getSchema(columnName);

		if (column.typeStr != TYPE_STR_COLOR) {
			throw new Error('Column is not color: ' + sheetName + '/' + columnName);
		}

		var value = this.getValue(columnName);
		var value16 = ('000000' + parseInt(value, 10).toString(16)).substr(-6);

		return [
			parseInt(value16.substr(0, 2), 16),
			parseInt(value16.substr(2, 2), 16),
			parseInt(value16.substr(4, 2), 16)
		];
	};

	CastleDBManagerItem.prototype.getCustomType = function(columnName) {
		assertParameters(arguments, ['columnName']);

		var definition = this.getCustomTypeDefinition(columnName);
		var value = this.getValue(columnName);

		var typeCase = definition[value[0]];
		var customType = {
			name: typeCase.name,
			args: {}
		};

		for (var i = 0; i < typeCase.args.length; i++) {
			var arg = typeCase.args[i];

			customType.args[arg.name] = value[i + 1];
		}

		return customType;
	};

	CastleDBManagerItem.prototype.getImage = function(columnName, hue) {
		assertParameters(arguments, ['columnName']);

		var column = this.getSchema(columnName);

		if (column.typeStr != TYPE_STR_IMAGE) {
			throw new Error('Column is not image: ' + sheetName + '/' + columnName);
		}

		if (this._manager._imageErrorUrl) {
			throw new Error('Failed to load: ' + this._manager._imageErrorUrl);
		}

		var value = this.getValue(columnName);
		var imageData = this._manager._image[value];

		if (imageData == null) {
			var address = sheetName + '/' + columnName;
			if (this._listIndex !== undefined) {
				address += '(' + this._listIndex + ')';
			}

			throw new Error('Image is not found: ' + address);
		}

		return ImageManager.loadCastleDBBitmap(
			this._manager._imagePath,
			value,
			imageData,
			hue || 0
		);
	};

	/**
	 * 必須パラメータのチェック
	 */
	var assertParameters = function(args, names) {
		for (var i = 0; i < names.length; i++) {
			if (names[i] == null) continue;

			if (args[i] === undefined) {
				throw new Error('missing required argument [' + names[i] + ']');
			}
		}
	};
})();
